/* utils.c - This file is a part of NutsOS
 *
 * NutsOS
 * Copyright (C) 2013  Free Software Foundation, Inc.
 *
 * NutsOS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * NutsOS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with NutsOS; If not, see <http://www.gnu.org/licenses/>.
 *
 * Valentin Verdier <valentin.verdier03@gmail.com>
 */

#include "utils.h"

#include "types.h"
#include "asm_x86.h"

void memcopy(char *src, char *dst, unsigned int n)
{
	unsigned int i;
	for(i = 0; i < n; i++)
	{
		dst[i] = src[i];
	}
}

int strcmp(char *str1, char *str2)
{
	unsigned int i;
	int length1 = 0;
	int length2 = 0;


	while(str1[length1] != 0)
	{
		length1++;
	}

	while(str2[length2] != 0)
	{
		length2++;
	}

	if(length1 == length2)
	{
		for(i = 0; i < length1; i++)
		{
			if(str1[i] != str2[i])
			{
				return 0;
			}
		}
	}
	else
	{
		return 0;
	}

	return 1;
}

void IntToChar(bits32 num, char *str, int base)
{
	const char digit[] = {'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F'};
	int i, j, a;
	int length;
	int tmp = 0;
	int counter = 0;

	for(i = 0; i < 10; i++)
	{
		if(num / expo(base, i) < base)
		{
			length = i;
			break;
		}
	}

	j = length;

	for(i = 0; i <= length; i++)
	{
		a = num / expo(base, j);
		str[counter] = digit[a - base * tmp];
		tmp = a;
		counter++;
		j--;
	}
	str[counter] = 0;
}

unsigned int expo(unsigned int x, unsigned int y)
{
	if(y > 0)
	{
		unsigned int i, res = x;
		for(i = 1; i < y; i++)
		{
			res *= x;
		}
		return res;
	}
	else
	{
		return 1;
	}
}

void LinkedListPushFront(void *addr, T_LinkedList *list)
{
	bits32 eflags = _SaveEflags();
	_ClearInterrupt();

	struct ElementModel *entry = addr;

	if(list->entryCount > 0)
	{
		InsertElement(0, entry, list->firstEntry);
	}
	else
	{
		InsertElement(0, entry, 0);
		list->lastEntry = entry;
	}
	list->firstEntry = entry;
	list->entryCount++;

	_RestoreEflags(eflags);
}

void LinkedListPopFront(T_LinkedList *list)
{
	bits32 eflags = _SaveEflags();
	_ClearInterrupt();

	struct ElementModel *nextEntry = list->firstEntry->next;
	if(list->entryCount > 1)
	{
		RemoveElement(list->firstEntry);
		list->firstEntry = nextEntry;
	}
	else
	{
		list->firstEntry = 0;
		list->lastEntry = 0;
	}
	list->entryCount--;

	_RestoreEflags(eflags);
}

void LinkedListPushBack(void *addr, T_LinkedList *list)
{
	bits32 eflags = _SaveEflags();
	_ClearInterrupt();

	struct ElementModel *entry = addr;

	if(list->entryCount > 0)
	{
		InsertElement(list->lastEntry, entry, 0);
	}
	else
	{
		InsertElement(0, entry, 0);
		list->firstEntry = entry;
	}
	list->lastEntry = entry;
	list->entryCount++;

	_RestoreEflags(eflags);
}

void LinkedListPopBack(T_LinkedList *list)
{
	bits32 eflags = _SaveEflags();
	_ClearInterrupt();

	struct ElementModel *previousEntry = list->lastEntry->prev;
	if(list->entryCount > 1)
	{
		RemoveElement(list->lastEntry);
		list->lastEntry = previousEntry;
	}
	else
	{
		list->firstEntry = 0;
		list->lastEntry = 0;
	}
	list->entryCount--;

	_RestoreEflags(eflags);
}

void LinkedListInsert(void *prev, void *addr, void *next, T_LinkedList *list)
{
	bits32 eflags = _SaveEflags();
	_ClearInterrupt();

	struct ElementModel *previousEntry = prev;
	struct ElementModel *entry = addr;
	struct ElementModel *nextEntry = next;

	if(nextEntry == list->firstEntry)
	{
		LinkedListPushFront(entry, list);
	}
	else if(previousEntry == list->lastEntry)
	{
		LinkedListPushBack(entry, list);
	}
	else
	{
		InsertElement(previousEntry, entry, nextEntry);
		list->entryCount++;
	}

	_RestoreEflags(eflags);
}

void LinkedListRemove(void *addr, T_LinkedList *list)
{
	bits32 eflags = _SaveEflags();
	_ClearInterrupt();

	struct ElementModel *entry = addr;

	if(entry == list->firstEntry)
	{
		LinkedListPopFront(list);
	}
	else if(entry == list->lastEntry)
	{
		LinkedListPopBack(list);
	}
	else
	{
		RemoveElement(entry);
		list->entryCount--;
	}

	_RestoreEflags(eflags);
}

void InsertElement(void *prev, void *addr, void *next)
{
	bits32 eflags = _SaveEflags();
	_ClearInterrupt();

	struct ElementModel *previousEntry = prev;
	struct ElementModel *entry = addr;
	struct ElementModel *nextEntry = next;

	if(previousEntry)
	{
		entry->prev = previousEntry;
		previousEntry->next = entry;
	}
	else
	{
		entry->prev = 0;
	}

	if(nextEntry)
	{
		entry->next = nextEntry;
		nextEntry->prev = entry;
	}
	else
	{
		entry->next = 0;
	}

	_RestoreEflags(eflags);
}

void RemoveElement(void *addr)
{
	bits32 eflags = _SaveEflags();
	_ClearInterrupt();

	struct ElementModel *entry = addr;
	struct ElementModel *previousEntry = entry->prev;
	struct ElementModel *nextEntry = entry->next;

	if(previousEntry)
	{
		if(nextEntry)
		{
			previousEntry->next = nextEntry;
		}
		else
		{
			previousEntry->next = 0;
		}
	}

	if(nextEntry)
	{
		if(previousEntry)
		{
			nextEntry->prev = previousEntry;
		}
		else
		{
			nextEntry->prev = 0;
		}
	}

	_RestoreEflags(eflags);
}

